@@ -38,6 +38,7 @@
 #include "goo-player-cd.h"
 #include "goo-player-info.h"
 #include "goo-window.h"
+#include "goo-volume-button.h"
 #include "gtk-utils.h"
 #include "glib-utils.h"
 #include "gconf-utils.h"
@@ -59,6 +60,8 @@
 #define DEFAULT_WIN_HEIGHT 400
 #define HIDE_TRACK_LIST "Hide _tracks"
 #define SHOW_TRACK_LIST "Show _tracks"
+#define DEFAULT_DEVICE "/dev/cdrom"
+#define DEFAULT_VOLUME 100
 
 struct _GooWindowPrivateData {
 	GtkUIManager    *ui;
@@ -72,6 +75,7 @@ struct _GooWindowPrivateData {
 	GtkWidget       *tray_popup_menu;
 
 	GtkWidget       *info;
+	GtkWidget       *volume_button;
 
 	GtkWidget       *tray;
 	GtkWidget       *tray_box;
@@ -164,15 +168,25 @@ window_update_statusbar_list_info (GooWindow *window)
 	}
 
 	if (tot_songs != 0) {
-		char  time_text[64];
-		char *info = NULL;
+		const char *album;
+		char        time_text[64];
+		char       *info = NULL;
+		char       *text;
+
+		album = goo_player_cd_get_album (GOO_PLAYER_CD (priv->player));
 
 		set_time_string (time_text, tot_time);
 		info = g_strdup_printf (ngettext ("%d track, %s", "%d tracks, %s", tot_songs), tot_songs, time_text);
 
-		gtk_statusbar_push (GTK_STATUSBAR (priv->statusbar), priv->list_info_cid, info);
+		if (album != NULL)
+			text = g_strconcat (album, ", ", info, NULL);
+		else
+			text = g_strconcat (info, NULL);
+
+		gtk_statusbar_push (GTK_STATUSBAR (priv->statusbar), priv->list_info_cid, text);
 
 		g_free (info);
+		g_free (text);
 	}
 }
 
@@ -237,6 +251,7 @@ window_update_sensitivity (GooWindow *window)
 	discid            = goo_player_cd_get_discid (GOO_PLAYER_CD (window->priv->player));
 
 	set_sensitive (window, "Play", !error && !playing);
+	set_sensitive (window, "PlaySelected", !error && one_file_selected);
 	set_sensitive (window, "Pause", !error && playing);
 	set_sensitive (window, "Stop", !error && (playing || paused));
 	set_sensitive (window, "TogglePlay", !error);
@@ -421,6 +436,12 @@ goo_window_finalize (GObject *object)
 	if (window->priv != NULL) {
 		GooWindowPrivateData *priv = window->priv;
 
+		/* Save preferences */
+		eel_gconf_set_integer (PREF_GENERAL_VOLUME, 
+				       goo_player_get_volume (priv->player));
+
+		/**/
+
 		gtk_object_destroy (GTK_OBJECT (priv->tooltips));
 		g_object_unref (priv->list_store);
 		
@@ -746,6 +767,7 @@ goo_window_class_init (GooWindowClass *class)
 	widget_class->realize   = goo_window_realize;
 	widget_class->unrealize = goo_window_unrealize;
 	widget_class->show      = goo_window_show;
+	widget_class->show      = goo_window_show;
 }
 
 
@@ -1142,9 +1164,13 @@ window_update_title (GooWindow *window)
 					 _("Paused"));
 
 	} else if (playing || (stopped && (priv->current_song != NULL))) {
-		title = g_strdup_printf ("%s - %s",
-					 priv->current_song->title,
-					 priv->current_song->artist);
+		if (priv->current_song->artist != NULL)
+			title = g_strdup_printf ("%s - %s",
+						 priv->current_song->title,
+						 priv->current_song->artist);
+		else
+			title = g_strdup_printf ("%s",
+						 priv->current_song->title);
 		
 	} else if (error) {
 		GError *error = goo_player_get_error (priv->player);
@@ -1419,12 +1445,55 @@ window_sync_ui_with_preferences (GooWindow *window)
 }
 
 
+static gboolean
+window_key_press_cb (GtkWidget   *widget, 
+		     GdkEventKey *event,
+		     gpointer     data)
+{
+	GooWindow *window = data;
+	gboolean   retval = FALSE;
+	int        new_song = -1;
+
+	if (window->priv->songs == 0)
+		return FALSE;
+
+	switch (event->keyval) {
+	case GDK_1:
+	case GDK_2:
+	case GDK_3:
+	case GDK_4:
+	case GDK_5:
+	case GDK_6:
+	case GDK_7:
+	case GDK_8:
+	case GDK_9:
+		new_song = event->keyval - GDK_1;
+		retval = TRUE;
+		break;
+	case GDK_0:
+		new_song = 10 - 1;
+		retval = TRUE;
+		break;
+	default:
+		break;
+	}
+
+	if (new_song >= 0 && new_song <= window->priv->songs - 1) {
+		goo_window_set_current_song (window, new_song);
+		goo_window_play (window);
+	}
+
+
+	return retval;
+}
+
+
 static void
 tray_object_destroyed (GooWindow *window)
 {
 	GooWindowPrivateData *priv = window->priv;
 
-	g_free (priv->tray_tips);
+	gtk_object_unref (GTK_OBJECT (priv->tray_tips));
 	priv->tray_tips = NULL;
 
 	if (priv->tray_popup_menu != NULL) {
@@ -1497,13 +1566,32 @@ tray_icon_expose (GtkWidget      *widget,
 }
 
 
+static void
+volume_button_changed_cb (GooVolumeButton *button,
+			  GooWindow       *window)
+{
+	int vol = (int) goo_volume_button_get_volume (button);
+	goo_player_set_volume (window->priv->player, vol);
+}
+
+
 static void
 goo_window_init (GooWindow *window)
 {
-	GooWindowPrivateData *priv;
+	window->priv = g_new0 (GooWindowPrivateData, 1);
+}
+
+
+static void
+goo_window_construct (GooWindow  *window,
+		      const char *location)
+{
+	GooWindowPrivateData *priv = window->priv;
 	GtkWidget            *menubar, *toolbar;
 	GtkWidget            *scrolled_window;
 	GtkWidget            *vbox;
+	GtkWidget            *hbox;
+	GtkWidget            *volume_box;
 	GtkWidget            *expander;
 	GtkTreeSelection     *selection;
 	int                   i;
@@ -1512,16 +1600,17 @@ goo_window_init (GooWindow *window)
 	GError               *error = NULL;		
 	char                 *device;
 
-	window->priv = g_new0 (GooWindowPrivateData, 1);
-	priv = window->priv;
-
 	priv->track_editor = CORBA_OBJECT_NIL;
 	
 	g_signal_connect (G_OBJECT (window), 
 			  "delete_event",
 			  G_CALLBACK (window_delete_event_cb),
 			  window);
-	
+	g_signal_connect (G_OBJECT (window), 
+			  "key_press_event",
+			  G_CALLBACK (window_key_press_cb), 
+			  window);
+
 	gnome_window_icon_set_from_default (GTK_WINDOW (window));
 
 	if (icon_size == 0) {
@@ -1703,13 +1792,31 @@ goo_window_init (GooWindow *window)
 
 	vbox = gtk_vbox_new (FALSE, 6);
 	gtk_container_set_border_width (GTK_CONTAINER (vbox), 6);
-	
+
+	hbox = gtk_hbox_new (FALSE, 12);
+	gtk_container_set_border_width (GTK_CONTAINER (hbox), 0);
+	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
+
 	priv->info = goo_player_info_new (NULL);
 	g_signal_connect (priv->info,
 			  "skip_to",
 			  G_CALLBACK (player_info_skip_to_cb), 
 			  window);
-	gtk_box_pack_start (GTK_BOX (vbox), priv->info, FALSE, FALSE, 0);
+	gtk_box_pack_start (GTK_BOX (hbox), priv->info, TRUE, TRUE, 0);
+
+	/**/
+
+	volume_box = gtk_vbox_new (FALSE, 0);
+	gtk_container_set_border_width (GTK_CONTAINER (volume_box), 0);
+	gtk_box_pack_start (GTK_BOX (hbox), volume_box, FALSE, FALSE, 0);
+
+	priv->volume_button = goo_volume_button_new (0.0, 100.0, 5.0);
+	gtk_box_pack_start (GTK_BOX (volume_box), priv->volume_button, FALSE, FALSE, 0);
+
+	g_signal_connect (priv->volume_button, 
+			  "changed",
+			  G_CALLBACK (volume_button_changed_cb), 
+			  window);
 
 	/**/
 
@@ -1729,6 +1836,7 @@ goo_window_init (GooWindow *window)
 	gtk_box_pack_start (GTK_BOX (vbox), 
 			    expander, 
 			    FALSE, FALSE, 0);
+
 	gtk_box_pack_start (GTK_BOX (vbox), scrolled_window, TRUE, TRUE, 0);
 
 	gnome_app_set_contents (GNOME_APP (window), vbox);
@@ -1744,7 +1852,10 @@ goo_window_init (GooWindow *window)
 
 	/**/
 
-	device = eel_gconf_get_string (PREF_GENERAL_DEVICE, NULL);
+	if (location != NULL)
+		device = g_strdup (location);
+	else
+		device = eel_gconf_get_string (PREF_GENERAL_DEVICE, DEFAULT_DEVICE);
 	priv->player = goo_player_cd_new (device);
 	g_free (device);
 
@@ -1771,6 +1882,10 @@ goo_window_init (GooWindow *window)
 
 	goo_player_info_set_player (GOO_PLAYER_INFO (priv->info), priv->player);
 
+	goo_volume_button_set_volume (GOO_VOLUME_BUTTON (priv->volume_button), 
+				      eel_gconf_get_integer (PREF_GENERAL_VOLUME, DEFAULT_VOLUME),
+				      TRUE);
+
 	/* Create the tray icon. */
 	
 	priv->tray = GTK_WIDGET (egg_tray_icon_new ("Goobox"));
@@ -1803,6 +1918,8 @@ goo_window_init (GooWindow *window)
 	
 	gtk_container_add (GTK_CONTAINER (priv->tray_box), priv->tray_icon);
 	priv->tray_tips = gtk_tooltips_new ();
+	gtk_object_ref (GTK_OBJECT (priv->tray_tips));
+	gtk_object_sink (GTK_OBJECT (priv->tray_tips));
 	gtk_tooltips_set_tip (GTK_TOOLTIPS (priv->tray_tips), 
 			      priv->tray_box, 
 			      _("CD Player"), 
@@ -1873,9 +1990,14 @@ goo_window_get_type ()
 
 
 GtkWindow * 
-goo_window_new (void)
+goo_window_new (const char *location)
 {
-	return GTK_WINDOW (g_object_new (GOO_TYPE_WINDOW, NULL));
+	GooWindow *window;
+
+	window = (GooWindow*) g_object_new (GOO_TYPE_WINDOW, NULL);
+	goo_window_construct (window, location);
+
+	return (GtkWindow*) window;
 }
 
 
@@ -1927,6 +2049,23 @@ goo_window_play (GooWindow *window)
 }
 
 
+void
+goo_window_play_selected (GooWindow *window)
+{
+	GList *song_list;
+
+	song_list = goo_window_get_song_list (window, TRUE);
+
+	if (g_list_length (song_list) == 1) {
+		SongInfo * song = song_list->data;
+		goo_window_set_current_song (window, song->number);
+		goo_window_play (window);
+	}
+
+	song_list_free (song_list);
+}
+
+
 void
 goo_window_stop (GooWindow *window)
 {
